import math
from soynlp.lemmatizer import conjugate
from soynlp.lemmatizer import _lemma_candidate

class StemExtractor:

    def __init__(self, lrgraph, stems, eomis, min_num_of_unique_R_char=10,
        min_entropy_of_R_char=0.5, min_entropy_of_R=1.5, verbose=True):

        self.lrgraph = lrgraph
        self.stems = stems
        self.eomis = eomis
        self.min_num_of_unique_R_char = min_num_of_unique_R_char
        self.min_entropy_of_R_char = min_entropy_of_R_char
        self.min_entropy_of_R = min_entropy_of_R
        self.verbose = verbose

        # L : stem surfaces
        # R : eomi surfaces
        self.L, self.R = self._conjugate_stem_and_eomi(lrgraph, stems, eomis)
        self._josa = {
            '거나', '게', '게는', '게도', '고',
            '고도', '고만', '는', '다', '다가',
            '서는', '아', '은'
        }

    def _print(self, message, replace=False, newline=True):
        header = '[Stem Extractor]'
        if replace:
            print('\r{} {}'.format(header, message),
                  end='\n' if newline else '', flush=True)
        else:
            print('{} {}'.format(header, message),
                  end='\n' if newline else '', flush=True)

    def _conjugate_stem_and_eomi(self, lrgraph, stems, eomis):
        eojeol_counter = lrgraph.to_EojeolCounter()

        stem_surfaces = set()
        eomi_surfaces = set()

        n_stems = len(stems)
        n_eomis = len(eomis)
        for i, stem in enumerate(stems):

            if self.verbose and i % 100 == 0:
                message = 'Checking combination of {} / {} stems + {} eomis'.format(
                    i, n_stems, n_eomis)
                self._print(message, replace=True, newline=False)

            stem_len = len(stem)
            for eomi in eomis:
                try:
                    for word in conjugate(stem, eomi):
                        if (eojeol_counter[word] == 0) or (len(word) <= stem_len):
                            continue
                        l, r = word[:stem_len], word[stem_len:]
                        stem_surfaces.add(l)
                        eomi_surfaces.add(r)
                except:
                    continue

        if self.verbose:
            message = 'Initializing was done with {} stems and {} eomis'.format(
                len(stems), len(eomis))
            self._print(message, replace=True, newline=True)

        del eojeol_counter
        return stem_surfaces, eomi_surfaces

    def extract(self, L_ignore=None, min_stem_score=0.7,
        min_frequency=100):

        if L_ignore is None:
            L_ignore = {}

        candidates = {}
        for r in self.R:
            for l, count in self.lrgraph.get_l(r, -1):
                if (l in self.L) or (l in L_ignore):
                    continue
                candidates[l] = candidates.get(l, 0) + count

        # 1st. frequency filtering
        candidates = {l:count for l, count in candidates.items()
            if count >= min_frequency}

        if self.verbose:
            message = 'batch prediction for {} candidates'.format(len(candidates))
            self._print(message, replace=False, newline=True)

        stem_surfaces = self._batch_prediction(
            candidates, min_stem_score, min_frequency)

        self.stem_surfaces, self.removals = self._post_processing(
            stem_surfaces)

        self.stems = self._to_stem(self.stem_surfaces)
        # TODO extracted 된 canonical form 이 다른 어미들과 조합되어 다르게 활용될 수 있는지 확인

        if self.verbose:
            message = '{} stems, {} surfacial stems, {} removals'.format(
                len(self.stems), len(self.stem_surfaces), len(self.removals))
            self._print(message, replace=False, newline=True)

        return self.stems

    def _batch_prediction(self, candidates,
        min_stem_score, min_frequency):

        # add known L for unknown L prediction
        extracted = {l:None for l in self.L}

        # from longer to shorter
        for l in sorted(candidates, key=lambda x:-len(x)):

            if ((l in self.L) or
                (l in self.R) or
                (len(l) == 1) or
                (l[-1] == '다') or # Hard coding rule
                (l in extracted)):
                continue

            score, freq = self.predict(l,
                min_stem_score, min_frequency)

            # no use entropy of R ?
            # entropy_of_R = _entropy([v for _, v in features])

            if (score < min_stem_score) or (freq < min_frequency):
                continue

            extracted[l] = (score, freq)

        # remove known L
        extracted = {l:score for l, score in extracted.items() if not (l in self.L)}

        return extracted

    def predict(self, l, min_stem_score=0.7, min_frequency=1, debug=False):

        features = self.lrgraph.get_r(l, -1)
        char_count = self._count_first_chars(features)

        unique_of_char = len(char_count)
        entropy_of_char = self._entropy(tuple(char_count.values()))

        pos, neg, unk = self._predict(l, features)
        score = (pos - neg) / (pos + neg) if (pos + neg) > 0 else 0
        freq = pos if score >= min_stem_score else neg + unk

        if debug:
            print('pos={}, neg={}, unk={}, n_features_={}, n_char={}, entropy_r={}'.format(
                pos, neg, unk, len(features), unique_of_char, entropy_of_char))

        if ((unique_of_char < self.min_num_of_unique_R_char) or
            (entropy_of_char < self.min_entropy_of_R_char)):
            return (0, 0)

        if freq < min_frequency:
            return (0, freq)
        else:
            return (score, freq)

    def _predict(self, l, features):
        pos, neg, unk = 0, 0, 0
        for r, freq in features:
            if r in self._josa: # 조사로도 이용되는 어미는 skip
                continue
            if not r:
                neg += freq
            elif r in self.R:
                pos += freq
            elif self._r_is_predicator(r):
                neg += freq
            elif self._exist_longer_eomi(l, r):
                neg += freq
            else:
                unk += freq
        return pos, neg, unk

    def _count_first_chars(self, features):
        char_count = [(r[0], count) for r, count in features if r]
        counter = {}
        for char, count in char_count:
            counter[char] = counter.get(char, 0) + count
        return counter

    def _entropy(self, counts):
        if len(counts) <= 1:
            return 0
        sum_ = sum(counts)
        entropy = [v/sum_ for v in counts]
        entropy = -1 * sum((p * math.log(p) for p in entropy))
        return entropy

    def _r_is_predicator(self, r):
        n = len(r)
        for i in range(1, n):
            if (r[:i] in self.L) and (r[i:] in self.R):
                return True
        return False

    def _exist_longer_eomi(self, l, r):
        for i in range(1, len(l)+1):
            if (l[-i:] + r) in self.R:
                return True
        return False

    # TODO check postprocessing rule
    def _post_processing(self, extracted):
        def is_stem_and_eomi(l):
            n = len(l)
            for i in range(1, n):
                if not ((l[:i] in self.L) or (l[:i] in extracted)):
                    continue
                for j in range(i+1, n+1):
                    if l[i:j] in self.R:
                        return True
            return False

        def exist_subword(l):
            for i in range(2, len(l)):
                if l[:i] in extracted:
                    return True
            return False

        removals = set()
        for l in sorted(extracted, key=lambda x:len(x)):
            if is_stem_and_eomi(l) or exist_subword(l):
                removals.add(l)
        extracted = {l:score for l, score in extracted.items()
                     if not (l in removals)}

        return extracted, removals

    def _to_stem(self, surfaces):

        def merge_score(freq0, score0, freq1, score1):
            return (freq0 + freq1, (score0 * freq0 + score1 * freq1) / (freq0 + freq1))

        stems = {}
        for l, (freq0, score0) in surfaces.items():
            for r, count in self.lrgraph.get_r(l, -1):
                try:
                    for stem, eomi in _lemma_candidate(l, r):
                        if eomi in self.eomis:
                            continue
                        stems[stem] = merge_score(
                            freq0, score0, *stems.get(stem, (0, 0)))
                except:
                    continue

        return stems


# -*- encoding:utf8 -*-

from soynlp.hangle import compose, decompose

def conjugate(stem, ending):

    assert ending # ending must be inserted

    l_len = len(stem)
    l_last = decompose(stem[-1])
    l_last_ = stem[-1]
    r_first = decompose(ending[0])
    r_first_ = compose(r_first[0], r_first[1], ' ') if r_first[1] != ' ' else ending[0]

    candidates = set()
    
    # ㄷ 불규칙 활용: 깨달 + 아 -> 깨달아
    if l_last[2] == 'ㄷ' and r_first[0] == 'ㅇ':
        l = stem[:-1] + compose(l_last[0], l_last[1], 'ㄹ')
        candidates.add(l + ending)

    # 르 불규칙 활용: 구르 + 어 -> 굴러
    if (l_last_ == '르') and (r_first_ == '아' or r_first_ == '어') and l_len >= 2:
        c0, c1, c2 = decompose(stem[-2])
        l = stem[:-2] + compose(c0, c1, 'ㄹ')
        r = compose('ㄹ', r_first[1], r_first[2]) + ending[1:]
        candidates.add(l + r)

    # ㅂ 불규칙 활용:
    # (모음조화) 더럽 + 어 -> 더러워 / 곱 + 아 -> 고와 
    # (모음조화가 깨진 경우) 아름답 + 아 -> 아름다워 / (-답, -꼽, -깝, -롭)
    if (l_last[2] == 'ㅂ') and (r_first_ == '어' or r_first_ == '아'):
        l = stem[:-1] + compose(l_last[0], l_last[1], ' ')
        if l_len >= 2 and (l_last_ == '답' or l_last_ == '곱' or l_last_ == '깝' or l_last_ == '롭'):
            c1 = 'ㅝ'
        elif r_first[1] == 'ㅗ':
            c1 = 'ㅘ'
        elif r_first[1] == 'ㅜ':
            c1 = 'ㅝ'
        elif r_first_ == '어':
            c1 = 'ㅝ'
        else: # r_first_ == '아'
            c1 = 'ㅘ'
        r = compose('ㅇ', c1, r_first[2]) + ending[1:]
        candidates.add(l + r)

    # 어미의 첫글자가 종성일 경우 (-ㄴ, -ㄹ, -ㅂ, -ㅆ)
    # 이 + ㅂ니다 -> 입니다
    if l_last[2] == ' ' and r_first[1] == ' ' and (r_first[0] == 'ㄴ' or r_first[0] == 'ㄹ' or r_first[0] == 'ㅂ' or r_first[0] == 'ㅆ'):
        l = stem[:-1] + compose(l_last[0], l_last[1], r_first[0])
        r = ending[1:]
        candidates.add(l + r)

    # ㅅ 불규칙 활용: 붓 + 어 -> 부어
    # exception : 벗 + 어 -> 벗어    
    if (l_last[2] == 'ㅅ') and (r_first[0] == 'ㅇ'):
        if stem[-1] == '벗':
            l = stem
        else:
            l = stem[:-1] + compose(l_last[0], l_last[1], ' ')
        candidates.add(l + ending)

    # 우 불규칙 활용: 푸 + 어 -> 퍼 / 주 + 어 -> 줘
    if l_last[1] == 'ㅜ' and l_last[2] == ' ' and r_first[0] == 'ㅇ' and r_first[1] == 'ㅓ':
        if l_last_ == '푸':
            l = '퍼'
        else:
            l = stem[:-1] + compose(l_last[0], 'ㅝ', r_first[2])
        r = ending[1:]
        candidates.add(l + r)

    # 오 활용: 오 + 았어 -> 왔어
    if l_last[1] == 'ㅗ' and l_last[2] == ' ' and r_first[0] == 'ㅇ' and r_first[1] == 'ㅏ':
        l = stem[:-1] + compose(l_last[0], 'ㅘ', r_first[2])
        r = ending[1:]
        candidates.add(l + r)

    # ㅡ 탈락 불규칙 활용: 끄 + 어 -> 꺼 / 트 + 었다 -> 텄다
    if (l_last_ == '끄' or l_last_ == '크' or l_last_ == '트') and (r_first[0] == 'ㅇ') and (r_first[1] == 'ㅓ'):
        l = stem[:-1] + compose(l_last[0], r_first[1], r_first[2])
        r = ending[1:]
        candidates.add(l + r)

    # 거라, 너라 불규칙 활용
    # '-거라/-너라'를 어미로 취급하면 규칙 활용
    if ending[:2] == '어라' or ending[:2] == '아라':
        if l_last[1] == 'ㅏ':            
            r = '거' + ending[1:]
        elif l_last[1] == 'ㅗ':
            r = '너' + ending[1:]
        else:
            r = ending
        candidates.add(stem + r)

    # 러 불규칙 활용: 이르 + 어 -> 이르러 / 이르 + 었다 -> 이르렀다
    if l_last_ == '르' and r_first[0] == 'ㅇ' and r_first[1] == 'ㅓ':
        r = compose('ㄹ', r_first[1], r_first[2]) + ending[1:]
        candidates.add(stem + r)

    # 여 불규칙 활용
    # 하 + 았다 -> 하였다 / 하 + 었다 -> 하였다
    if l_last_ == '하' and r_first[0] == 'ㅇ' and (r_first[1] == 'ㅏ' or r_first[1] == 'ㅓ'):
        # case 1
        r = compose(r_first[0], 'ㅕ', r_first[2]) + ending[1:]
        candidates.add(stem + r)
        # case 2
        l = stem[:-1] + compose('ㅎ', 'ㅐ', r_first[2])
        r = ending[1:]
        candidates.add(l + r)

    # ㅎ (탈락) 불규칙 활용
    # 파라 + 면 -> 파랗다 / 동그랗 + ㄴ -> 동그란
    if l_last[2] == 'ㅎ' and l_last_ != '좋' and not (r_first[1] == 'ㅏ' or r_first[1] == 'ㅓ'):
        if r_first[1] == ' ':
            l = l = stem[:-1] + compose(l_last[0], l_last[1], r_first[0])
        else:
            l = stem[:-1] + compose(l_last[0], l_last[1], ' ')
        if r_first_ == '으':
            r = ending[1:]
        elif r_first[1] == ' ':            
            r = ''
        else:
            r = ending
        candidates.add(l + r)

    # ㅎ (축약) 불규칙 할용
    # 파랗 + 았다 -> 파랬다 / 시퍼렇 + 었다 -> 시퍼렜다
    if l_last[2] == 'ㅎ' and l_last_ != '좋' and (r_first[1] == 'ㅏ' or r_first[1] == 'ㅓ'):
        l = stem[:-1] + compose(l_last[0], 'ㅐ' if r_first[1] == 'ㅏ' else 'ㅔ', r_first[2])
        r = ending[1:]
        candidates.add(l + r)

    # ㅎ + 네 불규칙 활용
    # ㅎ 탈락과 ㅎ 유지 모두 맞음
    if l_last[2] == 'ㅎ' and r_first[0] == 'ㄴ' and r_first[1] != ' ':
        candidates.add(stem + ending)

    # 이었 -> 였 규칙활용
    if ending[0] == '었' and l_last[1] == 'ㅣ' and l_last[2] == ' ':
        candidates.add(stem[:-1] + compose(l_last[0], 'ㅕ', 'ㅆ') + ending[1:])
        if l_last[0] == 'ㅇ':
            candidates.add(stem + ending)

    if not candidates and r_first[1] != ' ':
        candidates.add(stem + ending)

    return candidates

def _conjugate_stem(stem):

    l_len = len(stem)
    l_last = decompose(stem[-1])
    l_last_ = stem[-1]

    candidates = {stem}

    # ㄷ 불규칙 활용: 깨달 + 아 -> 깨달아
    if l_last[2] == 'ㄷ':
        l = stem[:-1] + compose(l_last[0], l_last[1], 'ㄹ')
        candidates.add(l)

    # 르 불규칙 활용: 구르 + 어 -> 굴러
    if (l_last_ == '르') and l_len >= 2:
        c0, c1, c2 = decompose(stem[-2])
        l = stem[:-2] + compose(c0, c1, 'ㄹ')
        candidates.add(l)

    # ㅂ 불규칙 활용:
    # (모음조화) 더럽 + 어 -> 더러워 / 곱 + 아 -> 고와
    # (모음조화가 깨진 경우) 아름답 + 아 -> 아름다워 / (-답, -꼽, -깝, -롭)
    if (l_last[2] == 'ㅂ'):
        l = stem[:-1] + compose(l_last[0], l_last[1], ' ')
        candidates.add(l)

    # 어미의 첫글자가 종성일 경우 (-ㄴ, -ㄹ, -ㅂ, -ㅆ)
    # 이 + ㅂ니다 -> 입니다
    if l_last[2] == ' ':
        candidates.add(stem[:-1] + compose(l_last[0], l_last[1], 'ㄴ'))
        candidates.add(stem[:-1] + compose(l_last[0], l_last[1], 'ㄹ'))
        candidates.add(stem[:-1] + compose(l_last[0], l_last[1], 'ㅂ'))
        candidates.add(stem[:-1] + compose(l_last[0], l_last[1], 'ㅆ'))

    # ㅅ 불규칙 활용: 붓 + 어 -> 부어
    # exception : 벗 + 어 -> 벗어
    if (l_last[2] == 'ㅅ') and stem[-1] != '벗':
        candidates.add(stem[:-1] + compose(l_last[0], l_last[1], ' '))

    # 우 불규칙 활용: 푸 + 어 -> 퍼 / 주 + 어 -> 줘
    if l_last[1] == 'ㅜ' and l_last[2] == ' ':
        if l_last_ == '푸':
            l = '퍼'
        else:
            candidates.add(stem[:-1] + compose(l_last[0], 'ㅝ', ' '))
            candidates.add(stem[:-1] + compose(l_last[0], 'ㅝ', 'ㅆ'))

    # 오 활용: 오 + 았어 -> 왔어
    if l_last[1] == 'ㅗ' and l_last[2] == ' ':
        candidates.add(stem[:-1] + compose(l_last[0], 'ㅘ', ' '))
        candidates.add(stem[:-1] + compose(l_last[0], 'ㅘ', 'ㅆ'))

    # ㅡ 탈락 불규칙 활용: 끄 + 어 -> 꺼 / 트 + 었다 -> 텄다
    if (l_last_ == '끄' or l_last_ == '크' or l_last_ == '트'):
        candidates.add(stem[:-1] + compose(l_last[0], 'ㅓ', ' '))
        candidates.add(stem[:-1] + compose(l_last[0], 'ㅓ', 'ㅆ'))

    # 거라, 너라 불규칙 활용
    # '-거라/-너라'를 어미로 취급하면 규칙 활용

    # 러 불규칙 활용: 이르 + 어 -> 이르러 / 이르 + 었다 -> 이르렀다

    # 여 불규칙 활용
    # 하 + 았다 -> 하였다 / 하 + 었다 -> 하였다
    # 하 + 았다 -> 했다
    if l_last_ == '하':
        candidates.add(stem[:-1] + '해')
        candidates.add(stem[:-1] + '했')

    # ㅎ (탈락) 불규칙 활용
    # 파라 + 면 -> 파랗다 / 동그랗 + ㄴ -> 동그란
    if l_last[2] == 'ㅎ' and l_last_ != '좋':
        candidates.add(stem[:-1] + compose(l_last[0], l_last[1], ' '))
        candidates.add(stem[:-1] + compose(l_last[0], l_last[1], 'ㄴ'))
        candidates.add(stem[:-1] + compose(l_last[0], l_last[1], 'ㄹ'))
        # candidates.add(stem[:-1] + compose(l_last[0], l_last[1], 'ㅂ'))
        candidates.add(stem[:-1] + compose(l_last[0], l_last[1], 'ㅆ'))

    # ㅎ (축약) 불규칙 할용
    # 파랗 + 았다 -> 파랬다 / 시퍼렇 + 었다 -> 시퍼렜다
    if l_last[2] == 'ㅎ' and l_last_ != '좋':
        candidates.add(stem[:-1] + compose(l_last[0], 'ㅐ', 'ㅆ'))
        # candidates.add(stem[:-1] + compose(l_last[0], 'ㅔ', 'ㅆ'))

    # ㅎ + 네 불규칙 활용
    # ㅎ 탈락과 ㅎ 유지 모두 맞음

    # 이었 -> 였 규칙활용
    if l_last[1] == 'ㅣ' and l_last[2] == ' ':
        candidates.add(stem[:-1] + compose(l_last[0], 'ㅕ', 'ㅆ'))

    return candidates


from collections import defaultdict
from collections import namedtuple
import os

from soynlp.utils import check_dirs
from soynlp.utils import EojeolCounter
from soynlp.utils import LRGraph
from soynlp.utils import get_process_memory
from soynlp.tokenizer import MaxScoreTokenizer
from ._josa import extract_domain_pos_features
from ._noun_postprocessing import detaching_features
from ._noun_postprocessing import ignore_features
from ._noun_postprocessing import check_N_is_NJ

NounScore = namedtuple('NounScore', 'frequency score')

class LRNounExtractor_v2:
    def __init__(self, max_left_length=10, max_right_length=9, predictor_headers=None,
        verbose=True, min_num_of_features=1, max_frequency_when_noun_is_eojeol=30,
        eojeol_counter_filtering_checkpoint=200000, min_eojeol_frequency=1,
        extract_compound=True, extract_pos_feature=False,
        extract_determiner=False, postprocessing=None, logpath=None):

        self.max_left_length = max_left_length
        self.max_right_length = max_right_length
        self.lrgraph = None
        self.verbose = verbose
        self.min_num_of_features = min_num_of_features
        self.max_frequency_when_noun_is_eojeol = max_frequency_when_noun_is_eojeol
        self.eojeol_counter_filtering_checkpoint = eojeol_counter_filtering_checkpoint
        self.min_eojeol_frequency = min_eojeol_frequency
        self.extract_compound = extract_compound
        self.extract_pos_feature = extract_pos_feature
        self.extract_determiner = extract_determiner
        self.logpath = logpath

        if logpath:
            check_dirs(logpath)

        if not postprocessing:
            postprocessing = [
                'detaching_features',
                'ignore_features',
                'ignore_NJ'
            ]
        elif isinstance(postprocessing) == str:
            postprocessing = [postprocessing]

        self.postprocessing = postprocessing

        if not predictor_headers:
            predictor_headers = self._set_default_predictor_header()

        self._load_predictor(predictor_headers)

    @property
    def is_trained(self):
        return self.lrgraph

    def _set_default_predictor_header(self):

        if self.verbose:
            print('[Noun Extractor] use default predictors')

        dirname = '/'.join(os.path.abspath(__file__).replace('\\', '/').split('/')[:-2])
        predictor_header = ['{}/trained_models/noun_predictor_ver2'.format(dirname)]

        return predictor_header

    def _load_predictor(self, headers):

        if type(headers) == str:
            headers = [headers]
        
        pos, neg = set(), set()
        for header in headers:

            # load positive features such as Josa
            pos_path = '{}_pos'.format(header)
            with open(pos_path, encoding='utf-8') as f:
                pos.update({feature.strip() for feature in f})

            # load negative features such as ending (Eomi)
            neg_path = '{}_neg'.format(header)
            with open(neg_path, encoding='utf-8') as f:
                neg.update({feature.strip() for feature in f})

        # common features such as -은 (조사/어미), -라고(조사/어미) 
        common = pos.intersection(neg)

        # remove common features from pos and neg
        pos = {feature for feature in pos if not (feature in common)}
        neg = {feature for feature in neg if not (feature in common)}

        if self.verbose:
            print('[Noun Extractor] num features: pos={}, neg={}, common={}'.format(
                len(pos), len(neg), len(common)))

        self._pos_features = pos
        self._neg_features = neg
        self._common_features = common

    def _append_features(self, feature_type, features):

        def check_feature_size():
            return (len(self._pos_features),
                    len(self._neg_features),
                    len(self._common_features))

        # size before
        n_pos, n_neg, n_common = check_feature_size()

        if feature_type == 'pos':
            commons = {f for f in features if (f in self._neg_features)}
            self._pos_features.update(
                {f for f in features if not (f in commons)})

        elif feature_type == 'neg':
            commons = {f for f in features if (f in self._pos_features)}
            self._neg_features.update(
                {f for f in features if not (f in commons)})

        elif feature_type == 'common':
            commons = features

        else:
            raise ValueError('Feature type was wrong. Choice = [pos, neg, common]')

        self._common_features.update(commons)

        # size after
        n_pos_, n_neg_, n_common_ = check_feature_size()

        if self.verbose:
            message = 'pos={} -> {}, neg={} -> {}, common={} -> {}'.format(
                n_pos, n_pos_, n_neg, n_neg_, n_common, n_common_)
            print('[Noun Extractor] features appended. {}'.format(message))

    def train_extract(self, sentences, min_noun_score=0.3,
        min_noun_frequency=1, min_eojeol_frequency=1, reset_lrgraph=True):

        self.train(sentences)

        return self.extract(min_noun_score, min_noun_frequency, reset_lrgraph)

    def train(self, sentences):

        if self.verbose:
            print('[Noun Extractor] counting eojeols')

        eojeol_counter = EojeolCounter(sentences, self.min_eojeol_frequency,
            max_length=self.max_left_length + self.max_right_length,
            filtering_checkpoint=self.eojeol_counter_filtering_checkpoint,
            verbose=self.verbose)

        self._num_of_eojeols = eojeol_counter._count_sum
        self._num_of_covered_eojeols = 0

        if self.verbose:
            print('[Noun Extractor] complete eojeol counter -> lr graph')

        self.lrgraph = eojeol_counter.to_lrgraph(
            self.max_left_length, self.max_right_length)

        if self.verbose:
            print('[Noun Extractor] has been trained. mem={} Gb'.format(
                '%.3f'%get_process_memory()))

    def _extract_determiner(self):
        raise NotImplemented

    def extract_domain_pos_features(self, append_extracted_features=True,
        noun_candidates=None, ignore_features=None,
        min_noun_score=0.3, min_noun_frequency=100,
        min_pos_score=0.3, min_pos_feature_frequency=1000,
        min_num_of_unique_lastchar=4, min_entropy_of_lastchar=0.5,
        min_noun_entropy=1.5):

        if self.verbose:
            print('[Noun Extractor] batch prediction for extracting pos feature')

        if not noun_candidates:
            noun_candidates = self._noun_candidates_from_positive_features()

        prediction_scores = self._batch_predicting_nouns(
            noun_candidates, min_noun_score)

        self.lrgraph.reset_lrgraph()

        self._pos_features_extracted = extract_domain_pos_features(
            prediction_scores,
            self.lrgraph,
            self._pos_features,
            ignore_features,
            min_noun_score,
            min_noun_frequency,
            min_pos_score,
            min_pos_feature_frequency,
            min_num_of_unique_lastchar,
            min_entropy_of_lastchar,
            min_noun_entropy
        )

        if append_extracted_features:
            self._append_features('pos', self._pos_features_extracted)

        if self.verbose:
            print('[Noun Extractor] {} pos features were extracted'.format(
                len(self._pos_features_extracted)))

    def extract(self, min_noun_score=0.3, min_noun_frequency=1, reset_lrgraph=True):

        # reset covered eojeol count
        self._num_of_covered_eojeols = 0

        # base prediction
        noun_candidates = self._noun_candidates_from_positive_features()

        if self.extract_pos_feature:
            if self.verbose:
                print('[Noun Extractor] extract and append pos features')

            self.extract_domain_pos_features(noun_candidates)

        prediction_scores = self._batch_predicting_nouns(
            noun_candidates, min_noun_score)

        if self.logpath:
            with open(self.logpath+'_prediction_score.log', 'w', encoding='utf-8') as f:
                f.write('noun score frequency\n')

                for word, score in sorted(prediction_scores.items(), key=lambda x:-x[1][1]):
                    f.write('{} {} {}\n'.format(word, score[0], score[1]))

        # E = N*J+ or N*Posi+
        if self.extract_compound:
            candidates = {l:sum(rdict.values()) for l,rdict in
                self.lrgraph._lr.items() if len(l) >= 4}
            compounds = self.extract_compounds(
                candidates, prediction_scores, min_noun_score)

        else:
            compounds = {}

        # combine single nouns and compounds
        nouns = {noun:score for noun, score in prediction_scores.items()
            if score[1] >= min_noun_score}

        nouns.update(compounds)

        # frequency filtering
        nouns = {noun:score for noun, score in nouns.items()
            if score[0] >= min_noun_frequency}

        nouns = self._post_processing(nouns, prediction_scores, compounds)

        if self.verbose:
            print('[Noun Extractor] {} nouns ({} compounds) with min frequency={}'.format(
                len(nouns), len(compounds), min_noun_frequency), flush=True)
            print('[Noun Extractor] flushing ... ', flush=True, end='')

        self._check_covered_eojeols(nouns)

        self._nouns = nouns

        if reset_lrgraph:
            # when extracting predicates, do not reset lrgraph.
            # the remained lrgraph is predicate (stem - ending) graph
            self.lrgraph.reset_lrgraph()

        nouns_ = {noun:NounScore(score[0], score[1]) for noun, score in nouns.items()}
        return nouns_

    def _get_nonempty_features(self, word, features):
        return [r for r, _ in features if (
            ( (r in self._pos_features) and (not self._exist_longer_pos(word, r)) ) or
            ( (r in self._neg_features) and (not self._exist_longer_neg(word, r)) ) )]

    def _exist_longer_pos(self, word, r):
        for e in range(len(word)-1, -1, -1):
            if (word[e:]+r) in self._pos_features:
                return True
        return False

    def _exist_longer_neg(self, word, r):
        for e in range(len(word)-1, -1, -1):
            if (word[e:]+r) in self._neg_features:
                return True
        return False

    def predict(self, word, min_noun_score=0.3, debug=False):

        # scoring
        features = self.lrgraph.get_r(word, -1)
        pos, common, neg, unk, end = self._predict(word, features)

        base = pos + neg
        score = 0 if base == 0 else (pos - neg) / base
        support = pos + end + common if score >= min_noun_score else neg + end + common

        features_ = self._get_nonempty_features(word, features)
        n_features_ = len(features_)

        # debug code
        if debug:
            print('pos={}, common={}, neg={}, unk={}, end={}, n_features_={}'.format(
                pos, common, neg, unk, end, n_features_))

        if n_features_ > self.min_num_of_features:
            return support, score

        else:
            # exception case
            sum_ = pos + common + neg + unk + end
            if sum_ == 0:
                return support, 0

            # exception. frequent nouns may have various positive R such as Josa
            if ((end > self.max_frequency_when_noun_is_eojeol) and (pos >= neg) ):
                return support, score

            if (common > 0 or pos > 0) and (end / sum_ >= 0.3) and (common >= neg):
                # 아이웨딩 + [('', 90), ('은', 3), ('측은', 1)] # 은 common / 대부분 단일어절 / 측은 unknown. 
                # 아이엠텍 + [('은', 2), ('', 2)]
                support = pos + common + end
                return (support, support / sum_)

            # 경찰국 + [(은, 1), (에, 1), (에서, 1)] -> {은, 에}
            first_chars = set()
            for r, _ in features:
                if not r:
                    continue
                if r in self._pos_features or r in self._common_features:
                    if not self._exist_longer_pos(word, r):
                        first_chars.add(r[0])
                if not (r in self._pos_features or r in self._common_features):
                    first_chars.add(r[0])

            if len(first_chars) >= 2:
                support = pos + common + end
                return (support, support / sum_)

            # Handling for post-processing in NounExtractor
            # Case 1.
            # 아이러브영주사과 -> 아이러브영주사 + [(과,1)] (minimum r feature 적용해야 하는 케이스) : 복합명사
            # 아이러브영주사과 + [('', 1)] 이므로, 후처리 이후 '아이러브영주사' 후보에서 제외됨
            # Case 2.
            # 아이였으므로 -> 아이였으므 + [(로, 2)] (minimum r feature 적용)
            # "명사 + Unknown R" 로 후처리
            return (support, 0)

    def _predict(self, word, features):

        pos, common, neg, unk, end = 0, 0, 0, 0, 0

        for r, freq in features:
            if r == '':
                end += freq
                continue
            if self._exist_longer_pos(word, r): # ignore
                continue
            if self._exist_longer_neg(word, r): # negative -다고
                neg += freq
                continue
            if r in self._common_features:
                common += freq
            elif r in self._pos_features:            
                pos += freq
            elif r in self._neg_features:
                neg += freq
            else:
                unk += freq

        return pos, common, neg, unk, end

    def _noun_candidates_from_positive_features(self, condition=None):

        def satisfy(word, e):
            return word[:e] == condition

        # noun candidates from positive featuers such as Josa
        N_from_J = {}

        for r in self._pos_features:
            for l, c in self.lrgraph.get_l(r, -1):

                # candidates filtering for debugging
                # condition is first chars in L
                if not condition:
                    N_from_J[l] = N_from_J.get(l,0) + c
                    continue

                # for debugging
                if not satisfy(l, len(condition)):
                    continue

                N_from_J[l] = N_from_J.get(l,0) + c

        return N_from_J

    def _batch_predicting_nouns(self,
        noun_candidates, min_noun_score=0.3):

        prediction_scores = {}

        n = len(noun_candidates)
        for i, word in enumerate(sorted(noun_candidates, key=lambda x:-len(x))):

            if self.verbose and i % 1000 == 999:
                percentage = '%.3f' % (100 * (i+1) / n)
                print('\r  -- batch prediction {} % of {} words'.format(
                    percentage, n), flush=True, end='')

            # base prediction
            support, score = self.predict(word, min_noun_score)
            prediction_scores[word] = (support, score)

            # if their score is higher than min_noun_score,
            # remove eojeol pattern from lrgraph
            if score >= min_noun_score:
                for r, count in self.lrgraph.get_r(word, -1):
                    # remove all eojeols that including word at left-side.
                    # we have to assume that pos, neg features are incomplete
                    self.lrgraph.remove_eojeol(word+r, count)
                    # if (r == '' or
                    #    (r in self._pos_features) or
                    #    (r in self._common_features)):
                    #    self.lrgraph.remove_eojeol(word+r, count)

        if self.verbose:
            print('\r[Noun Extractor] batch prediction was completed for {} words'.format(
                n), flush=True)

        return prediction_scores

    def extract_compounds(self, candidates, prediction_scores, min_noun_score=0.3):

        noun_scores = {noun:len(noun) for noun, score in prediction_scores.items()
                       if score[1] > min_noun_score and len(noun) > 1}

        self._compound_decomposer = MaxScoreTokenizer(scores=noun_scores)

        candidates = {l:sum(rdict.values()) for l,rdict in self.lrgraph._lr.items()
            if (len(l) >= 4) and not (l in noun_scores)}

        n = len(candidates)
        compounds_scores = {}
        compounds_counts = {}
        compounds_components = {}

        for i, (word, count) in enumerate(sorted(candidates.items(), key=lambda x:-len(x[0]))):

            if self.verbose and i % 1000 == 999:
                percentage = '%.2f' % (100 * i / n)
                print('\r  -- check compound {} %'.format(percentage), flush=True, end='')

            # skip if candidate is substring of longer compound
            if candidates.get(word, 0) <= 0:
                continue

            tokens = self._compound_decomposer.tokenize(word, flatten=False)[0]
            compound_parts = self._parse_compound(tokens)

            if compound_parts:

                # store compound components
                noun = ''.join(compound_parts)
                compounds_components[noun] = compound_parts

                # cumulate count and store compound score
                compound_score = max((prediction_scores.get(t, (0,0))[1] for t in compound_parts))
                compounds_scores[noun] = max(compounds_scores.get(noun,0), compound_score)
                compounds_counts[noun] = compounds_counts.get(noun,0) + count

                # reduce frequency of substrings
                for e in range(2, len(word)):
                    subword = word[:e]

                    if not subword in candidates:
                        continue

                    candidates[subword] = candidates.get(subword, 0) - count

                # eojeol coverage
                self.lrgraph.remove_eojeol(word)

        if self.verbose:
            print('\r[Noun Extractor] checked compounds. discovered {} compounds'.format(
                len(compounds_scores)))

        compounds = {noun:(score, compounds_counts.get(noun,0))
             for noun, score in compounds_scores.items()}

        self._compounds_components = compounds_components

        return compounds

    def decompose_compound(self, word):

        tokens = self._compound_decomposer.tokenize(word, flatten=False)[0]
        compound_parts = self._parse_compound(tokens)

        return (word, ) if not compound_parts else compound_parts

    def _parse_compound(self, tokens):
        """Check Noun* or Noun*Josa"""

        # format: (word, begin, end, score, length)
        for token in tokens[:-1]:
            if token[3] <= 0:
                return None

        # Noun* + Josa
        if len(tokens) >= 3 and tokens[-1][0] in self._pos_features:
            return tuple(t[0] for t in tokens[:-1])

        # all tokens are noun
        if tokens[-1][3] > 0:
            return tuple(t[0] for t in tokens)

        # else, not compound
        return None

    def _post_processing(self, nouns, prediction_scores, compounds):

        def print_status(method, nouns, removals):
            n_after = len(nouns)
            n_before = n_after + len(removals)
            print('[Noun Extractor] postprocessing {} : {} -> {}'.format(
                method, n_before, n_after))

        logpath = self.logpath+'_postprocessing.log' if self.logpath else None

        # initialize
        if logpath:
            with open(logpath, 'w', encoding='utf-8') as f:
                f.write('')

        for method in self.postprocessing:

            if method == 'detaching_features':

                logheader = '## Ignore noun candidates from detaching pos features\n'
                nouns, removals = detaching_features(nouns,
                    self._pos_features, logpath, logheader)

                if self.verbose:
                    print_status(method, nouns, removals)

            elif method == 'ignore_features':

                features = {f for f in self._pos_features}
                # features.update(self._neg_features)
                features.update(self._common_features)
                nouns, removals = ignore_features(nouns, features, logpath)

                if self.verbose:
                    print_status(method, nouns, removals)

            elif method == 'ignore_NJ':

                nouns, removals = check_N_is_NJ(nouns, self.lrgraph, logpath=logpath)

                if self.verbose:
                    print_status(method, nouns, removals)

        return nouns

    def _check_covered_eojeols(self, nouns):

        self.lrgraph.reset_lrgraph()

        noun_candidates = self._noun_candidates_from_positive_features()

        n = len(noun_candidates)
        for i, word in enumerate(sorted(noun_candidates, key=lambda x:-len(x))):

            if self.verbose and i % 1000 == 999:
                percentage = '%.3f' % (100 * (i+1) / n)
                print('\r[Noun Extractor] flushing ...  {} %'.format(
                    percentage), flush=True, end='')

            if not (word in nouns):
                continue

            if len(word) > 1:
                for r, count in self.lrgraph.get_r(word, -1):
                    # remove all eojeols that including word at left-side.
                    # we have to assume that pos, neg features are incomplete
                    self.lrgraph.remove_eojeol(word+r, count)
                    self._num_of_covered_eojeols += count
            else:
                # a syllable noun is exception; remove only N + pos feature
                if (r == '' or
                   (r in self._pos_features) or
                   (r in self._common_features)):
                    self.lrgraph.remove_eojeol(word+r, count)
                    self._num_of_covered_eojeols += count


        if self.verbose:
            print('\r[Noun Extractor] flushing was done. mem={} Gb{}'.format(
                '%.3f' % get_process_memory(), ' '*20), flush=True)
            coverage = '%.2f' % (100 * self._num_of_covered_eojeols
                / self._num_of_eojeols)
            print('[Noun Extractor] {} % eojeols are covered'.format(coverage), flush=True)


__title__ = 'soynlp'
__version__ = '0.0.49'
__author__ = 'Lovit'
__license__ = 'GPL v3'
__copyright__ = 'Copyright 2017 Lovit'

from . import hangle
from . import normalizer
from . import noun
from . import predicator
from . import pos
from . import tokenizer
from . import vectorizer
from . import word
from . import utils

# for compatibility
from .utils import DoublespaceLineCorpus

__all__ = [
    # modules
    'hangle',
    'normalizer',
    'noun',
    'predicator',
    'pos',
    'tokenizer',
    'vectorizer',
    'word',
    'utils',
    # for compatibility with ver <= 0.0.45
    'DoublespaceLineCorpus'
]


import os
from soynlp.utils import check_dirs

filepath = os.path.dirname(os.path.realpath(__file__))
josapath = filepath + '/frequent_enrolled_josa.txt'
suffixpath = filepath + '/frequent_noun_suffix.txt'

with open(josapath, encoding='utf-8') as f:
    josaset = {word.strip() for word in f if word}

with open(suffixpath, encoding='utf-8') as f:
    suffixset = {word.strip() for word in f if word}

def write_log(path, header, words):
    check_dirs(path)
    with open(path, 'a', encoding='utf-8') as f:
        f.write('\n{}\n'.format(header))
        for word in sorted(words):
            f.write('{}\n'.format(word))

def _select_true_nouns(nouns, removals):
    return {word:score for word, score in nouns.items()
            if (word in removals) == False}

def detaching_features(nouns, features, logpath=None, logheader=None):

    if not logheader:
        logheader = '## Ignored noun candidates from detaching features'

    removals = set()

    for word in nouns:

        if len(word) <= 2:
            continue

        for e in range(2, len(word)):

            l, r = word[:e], word[e:]

            # Skip a syllable word such as 고양이, 이력서
            if len(r) <= 1:
                continue

            if (l in nouns) and (r in features):
                removals.add(word)
                break

    if logpath:
        write_log(logpath, logheader, removals)

    nouns_ = _select_true_nouns(nouns, removals)
    return nouns_, removals

def ignore_features(nouns, features, logpath=None, logheader=None):

    if not logheader:
        logheader = '## Ignored noun candidates these are same with features'

    removals = set()

    for word in nouns:
        if word in features:
            removals.add(word)

    if logpath:
        write_log(logpath, logheader, removals)

    nouns_ = _select_true_nouns(nouns, removals)
    return nouns_, removals

def check_N_is_NJ(nouns, lrgraph, min_num_of_josa=5, logpath=None, logheader=None):

    if not logheader:
        logheader = '## Ignored true N+J'

    removals = set()
    for word, score in nouns.items():

        n = len(word)
        if n <= 2:
            continue

        for i in range(2, n):
            l, r = word[:i], word[i:]

            if (not (r in josaset)  # R 이 조사가 아니거나
                or (r in suffixset) # -서, -장, -이 처럼 suffix 이거나
                or not (l in nouns) # L 이 명사가 아니거나
                or score[0] >= nouns[l][0]): # L 의 명사 빈도수가 더 작으면
                continue

            features = lrgraph._lr_origin.get(l, {})
            features = [r for r in features if r in josaset]
            n_josa = len(features)
            if n_josa >= min_num_of_josa:
                removals.add(word)

    if logpath:
        write_log(logpath, logheader, removals)

    nouns_ = _select_true_nouns(nouns, removals)
    return nouns_, removals

def ngram_nouns(nouns, sents, min_count=10, min_score=0.7, max_n=2):

    ngram_nouns = {}
    return ngram_nouns


import soynlp
import setuptools
from setuptools import setup, find_packages


with open('README.md', encoding='utf-8') as fh:
    long_description = fh.read()

setuptools.setup(
    name="soynlp",
    version=soynlp.__version__,
    author=soynlp.__author__,
    author_email='soy.lovit@gmail.com',
    description="Unsupervised Korean Natural Language Processing Toolkits",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/lovit/soynlp',
    packages=setuptools.find_packages(),
    package_data={
        'soynlp':[
            'trained_models/*',
            'pos/dictionary/*.txt',
            'pos/dictionary/*/*.txt',
            'noun/*.txt',
            'lemmatizer/dictionary/default/*/*.txt'
        ]
    },
    keywords = [
        'korean-nlp',
        'korean-text-processing',
        'nlp',
        'tokenizer',
        'postagging',
        'word-extraction'
    ],
    install_requires=[
        "numpy>=1.12.1",
        "psutil>=5.0.1"
    ],
    classifiers=(
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ),
)


""" TERM DEFINITION
(l, r) : L and R position subwords
stem : stem of Adjective and Verb
ending : suffix, canonical form of ending

stems : set of stem including Adjectives and Verbs
composable_stems : stems that can be compounded with other prefix
    - [] + 하다 : 덕질+하다, 냐옹+하다, 냐옹+하냥
endings : set of ending
pos_l_features : canonical form set of stems (L subwords)
pos_composable_l_features : canonical form set of composable stems (L subwords)
lrgraph : L-R graph including [stem + Ending], Adverbs, 
          and maybe some Noun + Josa
"""

from collections import namedtuple
from soynlp.hangle import character_is_complete_korean
from soynlp.utils import LRGraph
from soynlp.utils import get_process_memory
from soynlp.utils import EojeolCounter
from soynlp.utils.utils import installpath
from soynlp.lemmatizer import _lemma_candidate
from soynlp.lemmatizer import _conjugate_stem
from ._eomi import EomiExtractor
from ._stem import StemExtractor

Predicator = namedtuple('Predicator', 'frequency lemma')

class PredicatorExtractor:

    def __init__(self, nouns, noun_pos_features=None, stems=None,
        eomis=None, extract_eomi=False, extract_stem=False, verbose=True):

        if not noun_pos_features:
            noun_pos_features = self._load_default_noun_pos_features()

        if not stems:
            stems = self._load_default_stems()

        if not eomis:
            eomis = self._load_default_eomis()

        self._nouns = nouns
        self._noun_pos_features = noun_pos_features
        self._stems = stems
        self._eomis = eomis
        self.verbose = verbose
        self.extract_eomi = extract_eomi
        self.extract_stem = extract_stem

        self._stem_surfaces = {l for stem in stems for l in _conjugate_stem(stem)}
        self.lrgraph = None

    def _load_default_noun_pos_features(self):
        path = '%s/trained_models/noun_predictor_ver2_pos' % installpath
        with open(path, encoding='utf-8') as f:
            pos_features = {word.split()[0] for word in f}
        return pos_features

    def _load_default_stems(self, min_frequency=100):
        dirs = '%s/lemmatizer/dictionary/default/Stem' % installpath
        paths = ['%s/Adjective.txt', '%s/Verb.txt']
        paths = [p % dirs for p in paths]
        stems = set()
        for path in paths:
            with open(path, encoding='utf-8') as f:
                for line in f:
                    word, frequency = line.split()
                    if int(frequency) < min_frequency:
                        continue
                    stems.add(word)
        return stems

    def _load_default_eomis(self, min_frequency=20):
        path = '%s/lemmatizer/dictionary/default/Eomi/Eomi.txt' % installpath
        eomis = set()
        with open(path, encoding='utf-8') as f:
            for line in f:
                word, frequency = line.split()
                if int(frequency) < min_frequency:
                    continue
                eomis.add(word)
        return eomis

    def _print(self, message, replace=False, newline=True):
        header = '[Predicator Extractor]'
        if replace:
            print('\r{} {}'.format(header, message),
                  end='\n' if newline else '', flush=True)
        else:
            print('{} {}'.format(header, message),
                  end='\n' if newline else '', flush=True)

    @property
    def is_trained(self):
        return self.lrgraph

    def train(self, sentences_or_lrgraph, min_eojeol_frequency=2,
        filtering_checkpoint=100000):

        if isinstance(sentences_or_lrgraph, LRGraph):
            self._train_with_lrgraph(sentences_or_lrgraph)
        else:
            self._train_with_sentences(sentences_or_lrgraph,
                min_eojeol_frequency, filtering_checkpoint)

    def _train_with_lrgraph(self, lrgraph):
        counter = {}
        for l, rdict in lrgraph._lr.items():
            for r, count in rdict.items():
                counter[l+r] = count

        self._num_of_eojeols = sum(counter.values())
        self._num_of_covered_eojeols = 0
        self.lrgraph = lrgraph

    def _train_with_sentences(self, sentences, min_eojeol_frequency=2,
        filtering_checkpoint=100000):

        check = filtering_checkpoint > 0

        if self.verbose:
            message = 'counting eojeols'
            self._print(message, replace=False, newline=False)

        # Eojeol counting
        counter = {}

        def contains_noun(eojeol, n):
            for e in range(2, n + 1):
                if eojeol[:e] in self._nouns:
                    return True
            return False

        for i_sent, sent in enumerate(sentences):

            if check and i_sent > 0 and i_sent % filtering_checkpoint == 0:
                counter = {
                    eojeol:count for eojeol, count in counter.items()
                    if count >= min_eojeol_frequency
                }

            if self.verbose and i_sent % 100000 == 99999:
                message = 'n eojeol = {} from {} sents. mem={} Gb{}'.format(
                    len(counter), i_sent + 1, '%.3f' % get_process_memory(), ' '*20)
                self._print(message, replace=True, newline=False)

            for eojeol in sent.split():

                n = len(eojeol)

                if n <= 1 or contains_noun(eojeol, n):
                    continue

                counter[eojeol] = counter.get(eojeol, 0) + 1

        if self.verbose:
            message = 'counting eojeols was done. {} eojeols, mem={} Gb{}'.format(
                len(counter), '%.3f' % get_process_memory(), ' '*20)
            self._print(message, replace=True, newline=True)

        counter = {
            eojeol:count for eojeol, count in counter.items()
            if count >= min_eojeol_frequency
        }

        self._num_of_eojeols = sum(counter.values())
        self._num_of_covered_eojeols = 0

        if self.verbose:
            message = 'complete eojeol counter -> lr graph'
            self._print(message, replace=False, newline=True)

        self.lrgraph = EojeolCounter()._to_lrgraph(
            counter,
            l_max_length=10,
            r_max_length=9
        )

        if self.verbose:
            message = 'has been trained. mem={} Gb'.format(
                '%.3f' % get_process_memory())
            self._print(message, replace=False, newline=True)

    def extract(self, candidates=None,
        min_predicator_frequency=10, reset_lrgraph=True,
        # Eomi extractor
        min_num_of_features=5, min_eomi_score=0.3, min_eomi_frequency=1,
        # Stem extractor
        min_num_of_unique_R_char=10, min_entropy_of_R_char=0.5,
        min_entropy_of_R=1.5, min_stem_score=0.7, min_stem_frequency=100):

        # reset covered eojeol count
        self._num_of_covered_eojeols = 0

        # TODO link parameters
        if self.extract_eomi:
            self._extract_eomi(min_num_of_features,
                min_eomi_score, min_eomi_frequency)

        # TODO link parameters
        if self.extract_stem:
            self._extract_stem(min_num_of_unique_R_char,
                min_entropy_of_R_char, min_entropy_of_R,
                min_stem_score, min_stem_frequency)

        return self._extract_predicator(
            candidates, min_predicator_frequency, reset_lrgraph)

    def _extract_eomi(self, min_num_of_features=5, min_eomi_score=0.3, min_eomi_frequency=1):

        eomi_extractor = EomiExtractor(
            lrgraph = self.lrgraph,
            stems = self._stems,
            nouns = self._nouns,
            min_num_of_features = min_num_of_features,
            verbose = self.verbose,
            logpath = None
        )

        extracted_eomis = eomi_extractor.extract(
            condition=None,
            min_eomi_score = min_eomi_score,
            min_frequency = min_eomi_frequency,
            reset_lrgraph=True
        )

        extracted_eomis = {eomi for eomi in extracted_eomis if not (eomi in self._eomis)}
        self._eomis.update(extracted_eomis)

        if self.verbose:
            message = '{} eomis have been extracted'.format(len(extracted_eomis))
            self._print(message, replace=False, newline=True)

    def _extract_stem(self, min_num_of_unique_R_char=10, min_entropy_of_R_char=0.5,
        min_entropy_of_R=1.5, min_stem_score=0.7, min_stem_frequency=100):

        stem_extractor = StemExtractor(
            lrgraph = self.lrgraph,
            stems = self._stems,
            eomis = self._eomis,
            min_num_of_unique_R_char = min_num_of_unique_R_char,
            min_entropy_of_R_char = min_entropy_of_R_char,
            min_entropy_of_R = min_entropy_of_R
        )

        extracted_stems = stem_extractor.extract(
            L_ignore=None,
            min_stem_score = min_stem_score,
            min_frequency = min_stem_frequency
        )

        extracted_stems = {stem for stem in extracted_stems if not (stem in self._stems)}
        self._stems.update(extracted_stems)

        if self.verbose:
            message = '{} stems have been extracted'.format(len(extracted_stems))
            self._print(message, replace=False, newline=True)

    def _extract_predicator(self, eojeols=None,
        min_frequency=10, reset_lrgraph=True):

        lemmas = self._as_lemma_candidates(
            eojeols, min_frequency)

        # TODO
        # evaluation

        if self.verbose:
            message = '{} predicators are extracted'.format(len(lemmas))
            self._print(message, replace=True, newline=True)

        return lemmas

    def _as_lemma_candidates(self, eojeols=None,  min_frequency=10):

        def all_character_are_complete_korean(s):
            for c in s:
                if not character_is_complete_korean(c):
                    return False
            return True

        if not eojeols:
            eojeols = {l:rdict.get('', 0) for l, rdict in self.lrgraph._lr.items()}
            eojeols = {eojeol:count for eojeol, count in eojeols.items()
                       if (count > min_frequency) and all_character_are_complete_korean(eojeol)}

        n_eojeols = len(eojeols)
        lemmas = {}

        for i_eojeol, eojeol in enumerate(eojeols):

            if self.verbose and i_eojeol % 5000 == 0:
                perc = '%.3f'% (100 * i_eojeol / n_eojeols)
                message = 'lemma candidates ... {} %'.format(perc)
                self._print(message, replace=True, newline=False)

            n = len(eojeol)
            lemma_candidates = set()

            for i in range(1, n+1):
                l, r = eojeol[:i], eojeol[i:]
                for stem, eomi in _lemma_candidate(l, r):
                    if (stem in self._stems) and (eomi in self._eomis):
                        lemma_candidates.add((stem, eomi))

            if lemma_candidates:
                count = eojeols[eojeol]
                lemmas[eojeol] = Predicator(count, lemma_candidates)
                self._num_of_covered_eojeols += count

        if self.verbose:
            message = 'lemma candidating was done. {} % eojeols are covered'.format(
                '%.3f' % (100 * self._num_of_covered_eojeols / self._num_of_eojeols))
            self._print(message, replace=True, newline=True)

        return lemmas


# -*- encoding:utf8 -*-

from collections import Counter
import numpy as np
from ._hangle import decompose

def levenshtein(s1, s2, cost={}):
    # based on Wikipedia/Levenshtein_distance#Python
    if len(s1) < len(s2):
        return levenshtein(s2, s1)

    if len(s2) == 0:
        return len(s1)
    
    def get_cost(c1, c2, cost):
        return 0 if (c1 == c2) else cost.get((c1, c2), 1)
    
    previous_row = range(len(s2) + 1)
    for i, c1 in enumerate(s1):
        current_row = [i + 1]
        for j, c2 in enumerate(s2):
            insertions = previous_row[j + 1] + 1 # j+1 instead of j since previous_row and current_row are one character longer
            deletions = current_row[j] + 1       # than s2
            substitutions = previous_row[j] + get_cost(c1, c2, cost)
            current_row.append(min(insertions, deletions, substitutions))
        previous_row = current_row
    
    return previous_row[-1]

def jamo_levenshtein(s1, s2):
    if len(s1) < len(s2):
        return jamo_levenshtein(s2, s1)

    if len(s2) == 0:
        return len(s1)
    
    def get_jamo_cost(c1, c2):
        return 0 if (c1 == c2) else levenshtein(decompose(c1), decompose(c2))/3

    previous_row = range(len(s2) + 1)
    for i, c1 in enumerate(s1):
        current_row = [i + 1]
        for j, c2 in enumerate(s2):
            insertions = previous_row[j + 1] + 1 # j+1 instead of j since previous_row and current_row are one character longer
            deletions = current_row[j] + 1       # than s2
            substitutions = previous_row[j] + get_jamo_cost(c1, c2)
            current_row.append(min(insertions, deletions, substitutions))
        previous_row = current_row
    
    return previous_row[-1]

def cosine_distance(s1, s2, unitfy=lambda x:Counter(x)):
    '''distance = 1 - cosine similarity; [0, 2] '''
    if (not s1) or (not s2):
        return 2
    
    d1 = unitfy(s1)
    d2 = unitfy(s2)
    prod = 0
    for c1, f in d1.items():
        prod += (f * d2.get(c1, 0))
    return 1 - ( prod / np.sqrt( (sum([f**2 for f in d1.values()]) * sum([f**2 for f in d2.values()])) ) )

def jaccard_distance(s1, s2, unitfy=lambda x:set(x)):
    if (not s1) or (not s2):
        return 1
    
    s1_set = unitfy(s1)
    s2_set = unitfy(s2)
    return 1 - len(s1_set.intersection(s2_set)) / len(s1_set.union(s2_set))



from ._dictionary import Dictionary
from ._evaluator import BaseEvaluator, SimpleEojeolEvaluator, LREvaluator
from ._pos_extractor import POSExtractor
from ._template import BaseTemplateMatcher, EojeolTemplateMatcher, LRTemplateMatcher
from ._tagger import BaseTagger, SimpleTagger
from ._tagger import BasePostprocessor, UnknowLRPostprocessor
from . import tagset
#from ._lrtagger import LRMaxScoreTagger


# -*- encoding:utf8 -*-

from collections import defaultdict
import sys
from soynlp.utils import get_process_memory


class EojeolPatternTrainer:

    def __init__(self, max_left_length=10, max_right_length=6, min_frequency=10, verbose=True):
        self.max_left_length = max_left_length
        self.max_right_length = max_right_length
        self.min_frequency = min_frequency
        self.verbose = verbose
        self.lrgraph = None
        self.rlgraph = None
        self.wordset_l = None
        self.wordset_r = None
    
    def train(self, sents, wordset_l=None, wordset_r=None):
        if (not wordset_l) or (not wordset_r):
            wordset_l, wordset_r = self._scan_vocabulary(sents)
        self.lrgraph, self.rlgraph = self._build_graph(sents, wordset_l, wordset_r)
        # TODO more
        
    def _scan_vocabulary(self, sents):
        """
        Parameters
        ----------
            sents: list-like iterable object which has string
            
        It computes subtoken frequency first. 
        After then, it builds lr-graph with sub-tokens appeared at least min count
        """
        
        _ckpt = int(len(sents) / 40)
        
        wordset_l = defaultdict(lambda: 0)
        wordset_r = defaultdict(lambda: 0)
        
        for i, sent in enumerate(sents):
            for token in sent.split(' '):
                if not token:
                    continue
                token_len = len(token)
                for i in range(1, min(self.max_left_length, token_len)+1):
                    wordset_l[token[:i]] += 1
                for i in range(1, min(self.max_right_length, token_len)):
                    wordset_r[token[-i:]] += 1
            if self.verbose and (i % _ckpt == 0):
                args = ('#' * int(i/_ckpt), '-' * (40 - int(i/_ckpt)), 100.0 * i / len(sent), '%', get_process_memory())
                sys.stdout.write('\rscanning: %s%s (%.3f %s) %.3f Gb' % args)
            
        wordset_l = {w for w,f in wordset_l.items() if f >= self.min_frequency}
        wordset_r = {w for w,f in wordset_r.items() if f >= self.min_frequency}
        if self.verbose:
            print('\rscanning completed')
            print('(L,R) has (%d, %d) tokens. memory = %.3f Gb' % (len(wordset_l), len(wordset_r), get_process_memory()))

        return wordset_l, wordset_r
    
    def _build_graph(self, sents, wordset_l, wordset_r):
        self.wordset_l = wordset_l
        self.wordset_r = wordset_r
        self.wordset_r.add('')
        _ckpt = int(len(sents) / 40)
        lrgraph = defaultdict(lambda: defaultdict(lambda: 0))
        rlgraph = defaultdict(lambda: defaultdict(lambda: 0))
        
        for i, sent in enumerate(sents):
            for token in sent.split():
                if not token:
                    continue
                token_len = len(token)
                for i in range(1, min(self.max_left_length, token_len)+1):
                    l = token[:i]
                    r = token[i:]
                    if (not l in wordset_l) or (not r in wordset_r):
                        continue
                    lrgraph[l][r] += 1
                    rlgraph[r][l] += 1

            if self.verbose and (i % _ckpt == 0):
                args = ('#' * int(i/_ckpt), '-' * (40 - int(i/_ckpt)), 100.0 * i / len(sents), '%', get_process_memory())
                sys.stdout.write('\rbuilding lr-graph: %s%s (%.3f %s), memory = %.3f Gb' % args)
        if self.verbose:                       
            sys.stdout.write('\rbuilding lr-graph completed. memory = %.3f Gb' % get_process_memory())
            
        lrgraph = {l:{r:f for r,f in rdict.items()} for l,rdict in lrgraph.items()}
        rlgraph = {r:{l:f for l,f in ldict.items()} for r, ldict in rlgraph.items()}
        return lrgraph, rlgraph
    
    def save(self, fname):
        with open(fname, 'w', encoding='utf-8') as f:
            f.write('%d %d %d %d\n' % (self.max_left_length, self.max_right_length, self.min_frequency, 1 if self.verbose else 0))
            f.write('# lrgraph\n')
            for l, rdict in self.lrgraph.items():
                f.write('> %s (%d)\n' % (l, sum(rdict.values())))
                for r, freq in sorted(rdict.items(), key=lambda x:x[1], reverse=True):
                    f.write('  - %s: %d\n' % (r, freq))
            f.write('\n# rlgraph\n')
            for r, ldict in self.rlgraph.items():
                f.write('> %s (%d)\n' % (r, sum(ldict.values())))
                for l, freq in sorted(ldict.items(), key=lambda x:x[1], reverse=True):
                    f.write('  - %s: %d\n' % (l, freq))
            
    def load(self, fname):
        with open(fname, encoding='utf-8') as f:
            param = next(f).strip()
            args = param.split()
            try:
                args = [int(a) for a in args]
                if len(args) != 4:
                    raise ValueError('first line should be parameter info, %s' % str(e))
                self.max_left_length = args[0]
                self.max_right_length = args[1]
                self.min_frequency = args[2]
                self.verbose == True if args[3] == 1 else False
            except Exception as e:
                raise ValueError('first line should be parameter info, %s' % str(e))
                
            lrgraph = defaultdict(lambda: defaultdict(lambda: 0))
            rlgraph = defaultdict(lambda: defaultdict(lambda: 0))

            load_type = next(f).strip()
            if not load_type == '# lrgraph':
                raise ValueError('Cannot find lrgraph data, %s' % load_type)
                
            key1 = None
            for row in f:
                row = row[:-1]
                if not row: continue
                if row == '# rlgraph':break
                
                if row[:2] == '> ':
                    key1 = row[2:row.rindex('(')].strip()
                    continue
                if row[:4] == '  - ':
                    key2, freq = row[4:].split(': ')
                    freq = int(freq)
                    lrgraph[key1][key2] = freq
            self.wordset_l = set(lrgraph.keys())
                
            for row in f:
                row = row[:-1]
                if not row: continue
                if row[:2] == '> ':
                    key1 = row[2:row.rindex('(')].strip()
                    continue
                if row[:4] == '  - ':
                    key2, freq = row[4:].split(': ')
                    freq = int(freq)
                    rlgraph[key1][key2] = freq
            self.wordset_r = set(rlgraph.keys())
            
            lrgraph = {l:{r:f for r,f in rdict.items()} for l,rdict in lrgraph.items()}
            rlgraph = {r:{l:f for l,f in ldict.items()} for r, ldict in rlgraph.items()}
            self.lrgraph = lrgraph
            self.rlgraph = rlgraph
            
    def train_hits(lrgraph=None, rlgraph=None, sum_of_rank=10000, decaying_factor=0.9, max_iter=10, tolerance=0.0001):
        def normalize(g, sum_of_rank, df):
            factor = df * sum_of_rank / sum(g.values())
            restart = (1 - df) * sum_of_rank / len(g)
            g_ = {word:(factor*rank + restart) for word, rank in g.items() if word != ''}
            return g_

        if lrgraph == None:
            (lrgraph, rlgraph) = (self.lrgraph, self.rlgraph)
        
        rank = sum_of_rank / len(lrgraph)
        rank_l = {l:rank for l in lrgraph.keys()}
        rank = sum_of_rank / len(rlgraph)
        rank_r = {r:rank for r in rlgraph.keys() if r != ''}

        for n_iter in range(max_iter):
            next_rank_l = {}
            for l, rdict in lrgraph.items():
                sum_rrank = sum({freq*rank_r[r] for r, freq in rdict.items() if r != ''})
                next_rank_l[l] = sum_rrank
            next_rank_l = normalize(next_rank_l, sum_of_rank, decaying_factor)

            next_rank_r = {}
            for r, ldict in rlgraph.items():
                if r == '': continue
                sum_lrank = sum({freq*rank_l[l] for l, freq in ldict.items()})
                next_rank_r[r] = sum_lrank
            next_rank_r = normalize(next_rank_r, sum_of_rank, decaying_factor)

            if self.verbose:
                sys.stdout.write('\rtrain hits ... %d in %d' % (n_iter+1, max_iter))

            diff = sum([abs(rank - next_rank_l[w]) for w, rank in rank_l.items()])
            diff += sum([abs(rank - next_rank_r[w]) for w, rank in rank_r.items()])
            rank_l = next_rank_l
            rank_r = next_rank_r
            if diff < (sum_of_rank * tolerance):
                if self.verbose:
                    print('\rgraph was converged at %d iteration' % (n_iter+1))
                break

        if self.verbose:
            print('\rcomputation was done at %d iteration' % (n_iter+1))

        return rank_l, rank_r



# -*- encoding:utf8 -*-

import sys
if sys.version_info <= (2,7):
    reload(sys)
    sys.setdefaultencoding('utf-8')
import warnings
import re
import numpy as np

kor_begin     = 44032
kor_end       = 55203
chosung_base  = 588
jungsung_base = 28
jaum_begin = 12593
jaum_end = 12622
moum_begin = 12623
moum_end = 12643

chosung_list = [ 'ㄱ', 'ㄲ', 'ㄴ', 'ㄷ', 'ㄸ', 'ㄹ', 'ㅁ', 'ㅂ', 'ㅃ', 
        'ㅅ', 'ㅆ', 'ㅇ' , 'ㅈ', 'ㅉ', 'ㅊ', 'ㅋ', 'ㅌ', 'ㅍ', 'ㅎ']

jungsung_list = ['ㅏ', 'ㅐ', 'ㅑ', 'ㅒ', 'ㅓ', 'ㅔ', 
        'ㅕ', 'ㅖ', 'ㅗ', 'ㅘ', 'ㅙ', 'ㅚ', 
        'ㅛ', 'ㅜ', 'ㅝ', 'ㅞ', 'ㅟ', 'ㅠ', 
        'ㅡ', 'ㅢ', 'ㅣ']

jongsung_list = [
    ' ', 'ㄱ', 'ㄲ', 'ㄳ', 'ㄴ', 'ㄵ', 'ㄶ', 'ㄷ',
        'ㄹ', 'ㄺ', 'ㄻ', 'ㄼ', 'ㄽ', 'ㄾ', 'ㄿ', 'ㅀ', 
        'ㅁ', 'ㅂ', 'ㅄ', 'ㅅ', 'ㅆ', 'ㅇ', 'ㅈ', 'ㅊ', 
        'ㅋ', 'ㅌ', 'ㅍ', 'ㅎ']

jaum_list = ['ㄱ', 'ㄲ', 'ㄳ', 'ㄴ', 'ㄵ', 'ㄶ', 'ㄷ', 'ㄸ', 'ㄹ', 
              'ㄺ', 'ㄻ', 'ㄼ', 'ㄽ', 'ㄾ', 'ㄿ', 'ㅀ', 'ㅁ', 'ㅂ', 
              'ㅃ', 'ㅄ', 'ㅅ', 'ㅆ', 'ㅇ', 'ㅈ', 'ㅉ', 'ㅊ', 'ㅋ', 'ㅌ', 'ㅍ', 'ㅎ']

moum_list = ['ㅏ', 'ㅐ', 'ㅑ', 'ㅒ', 'ㅓ', 'ㅔ', 'ㅕ', 'ㅖ', 'ㅗ', 'ㅘ', 
              'ㅙ', 'ㅚ', 'ㅛ', 'ㅜ', 'ㅝ', 'ㅞ', 'ㅟ', 'ㅠ', 'ㅡ', 'ㅢ', 'ㅣ']

doublespace_pattern = re.compile('\s+')
repeatchars_pattern = re.compile('(\w)\\1{3,}')

def normalize(doc, english=False, number=False, punctuation=False, remove_repeat = 0, remains={}):
    message = 'normalize func will be moved soynlp.normalizer at ver 0.1'
    warnings.warn(message, DeprecationWarning, stacklevel=2)

    if remove_repeat > 0:
        doc = repeatchars_pattern.sub('\\1' * remove_repeat, doc)

    f = []    
    for c in doc:
        if (c == ' '):
            f.append(c)
            continue
        i = to_base(c)
        if (kor_begin <= i <= kor_end) or (jaum_begin <= i <= jaum_end) or (moum_begin <= i <= moum_end):
            f.append(c)
            continue
        if (english) and ((i >= 97 and i <= 122) or (i >= 65 and i <= 90)):
            f.append(c)
            continue
        if (number) and (i >= 48 and i <= 57):
            f.append(c)
            continue
        if (punctuation) and (i == 33 or i == 34 or i == 39 or i == 44 or i == 46 or i == 63 or i == 96):
            f.append(c)
            continue
        if c in remains:
            f.append(c)
            continue
        else:
            f.append(' ')            
    return doublespace_pattern.sub(' ', ''.join(f)).strip()

def compose(chosung, jungsung, jongsung):
    return chr(kor_begin + chosung_base * chosung_list.index(chosung) + jungsung_base * jungsung_list.index(jungsung) + jongsung_list.index(jongsung))

def decompose(c):
    if not character_is_korean(c):
        return None
    i = to_base(c)
    if (jaum_begin <= i <= jaum_end):
        return (c, ' ', ' ')
    if (moum_begin <= i <= moum_end):
        return (' ', c, ' ')    
    i -= kor_begin
    cho  = i // chosung_base
    jung = ( i - cho * chosung_base ) // jungsung_base 
    jong = ( i - cho * chosung_base - jung * jungsung_base )    
    return (chosung_list[cho], jungsung_list[jung], jongsung_list[jong])

def character_is_korean(c):
    i = to_base(c)
    return (kor_begin <= i <= kor_end) or (jaum_begin <= i <= jaum_end) or (moum_begin <= i <= moum_end)

def character_is_complete_korean(c):
    return (kor_begin <= to_base(c) <= kor_end)

def character_is_jaum(c):
    return (jaum_begin <= to_base(c) <= jaum_end)

def character_is_moum(c):
    return (moum_begin <= to_base(c) <= moum_end)

def to_base(c):
    if sys.version_info.major == 2:
        if type(c) == str or type(c) == unicode:
            return ord(c)
        else:
            raise TypeError
    else:
        if type(c) == str or type(c) == int:
            return ord(c)
        else:
            raise TypeError

def character_is_number(i):
    i = to_base(i)
    return (i >= 48 and i <= 57)

def character_is_english(i):
    i = to_base(i)
    return (i >= 97 and i <= 122) or (i >= 65 and i <= 90)

def character_is_punctuation(i):
    i = to_base(i)
    return (i == 33 or i == 34 or i == 39 or i == 44 or i == 46 or i == 63 or i == 96)

class ConvolutionHangleEncoder:
    """초/중/종성을 구성하는 자음/모음과 띄어쓰기만 인코딩
    one hot vector [ㄱ, ㄴ, ㄷ, ... ㅎ, ㅏ, ㅐ, .. ㅢ, ㅣ,"  ", ㄱ, ㄲ, ... ㅍ, ㅎ,"  ", 0, 1, 2, .. 9]
    """
    def __init__(self):
        self.jung_begin = 19 # len(chosung_list)
        self.jong_begin = 40 # self.jung_begin + len(jungsung_list)
        self.number_begin = 68 # self.jong_begin + len(jongsung_list)
        self.space = 78 # len(chosung_list) + len(jungsung_list) + len(jongsung_list) + 10
        self.dim = 79
    
    def encode(self, sent):
        onehot = self.sent_to_onehot(sent)
        x = np.zeros((len(onehot), self.dim))
        for i, xi in enumerate(onehot):
            for j in xi:
                x[i,j] = 1
        return x
    
    def sent_to_onehot(self, sent):
        sent = self._normalize(sent)
        sent = [ord(c) for c in sent]
        sent_ = []
        for i in sent:
            if i == 32: sent_.append((self.space,))
            elif 48 <= i <= 57: sent_.append((i - 48 + self.number_begin, ))
            else: sent_.append(self._decompose(i))
        return sent_
    
    def onehot_to_sent(self, encoded_sent):
        chars = []
        for c in encoded_sent:
            if len(c) == 1:
                idx = c[0] - self.number_begin
                if idx == 10:
                    chars.append(' ')
                elif 0 <= idx < 10:
                    chars.append(str(idx))
                else:
                    chars.append(None)
            elif len(c) == 3:
                cho, jung, jong = c
                if (0 <= cho < self.jung_begin) and (self.jung_begin <= jung < self.jong_begin) and (self.jong_begin <= jong < self.number_begin):
                    chars.append(compose(chosung_list[cho], jungsung_list[jung - self.jung_begin], jongsung_list[jong - self.jong_begin]))
                else:
                    chars.append(None)
            else:
                chars.append(None)
        return ''.join(chars)
        
    def _normalize(self, sent):
        import re
        regex = re.compile('[^ㄱ-ㅎㅏ-ㅣ가-힣 0-9]')
        sent = regex.sub(' ', sent)
        sent = doublespace_pattern.sub(' ', sent)
        return sent
        
    def _compose(self, cho, jung, jong):
        return chr(kor_begin + chosung_base * cho + jungsung_base * jung + jong)

    def _decompose(self, i):
        if (jaum_begin <= i <= jaum_end):
            return (i - jaum_begin,)
        if (moum_begin <= i <= moum_end):
            return (i - moum_begin,)
        i -= kor_begin
        cho  = i // chosung_base
        jung = ( i - cho * chosung_base ) // jungsung_base 
        jong = ( i - cho * chosung_base - jung * jungsung_base )    
        return (cho, self.jung_begin + jung, self.jong_begin + jong)